import json
from typing import Optional

from jwskate import JweCompact, Jwk, JwtSigner

from bas_metadata_library.standards.magic_administration.v1 import AdministrationMetadata


class AdministrationMetadataSubjectMismatchError(Exception):
    """Raised when administration metadata does not relate to discovery metadata for the same resource."""

    pass


class AdministrationMetadataIntegrityError(Exception):
    """Raised when administration metadata ID does not match the subject of a JWT."""

    pass


class AdministrationKeys:
    """Encryption and signing keys for administration metadata."""

    def __init__(
        self, encryption_private: Jwk, signing_public: Optional[Jwk] = None, signing_private: Optional[Jwk] = None
    ) -> None:
        """One of the public or private signing key is needed depending on whether metadata needs verification and/or signing."""
        self.encryption_private: Jwk = encryption_private
        self.signing_private: Optional[Jwk] = signing_private

        if isinstance(signing_public, Jwk):
            self.signing_public: Jwk = signing_public
            return
        if isinstance(signing_private, Jwk):
            self.signing_public: Jwk = signing_private.public_jwk()
            return

        msg = "Public or private signing_key must be provided."
        raise TypeError(msg) from None

    def __getstate__(self) -> dict[str, str]:
        """
        Support pickling by dumping keys to JSON.

        Crytography keys within JWKs cannot be pickled.
        """
        out = {"encryption_private": self.encryption_private.to_json(), "signing_public": self.signing_public.to_json()}
        if self.signing_private:
            out["signing_private"] = self.signing_private.to_json()
        return out

    def __setstate__(self, state: dict[str, str]) -> None:
        """Load keys from JSON when unpickling."""
        self.encryption_private = Jwk.from_json(state["encryption_private"])
        self.signing_public = Jwk.from_json(state["signing_public"])
        try:
            self.signing_private = Jwk.from_json(state["signing_private"])
        except KeyError:
            self.signing_private = None

    def __eq__(self, other: object) -> bool:
        """Check equality."""
        if not isinstance(other, AdministrationKeys):
            return NotImplemented
        return (
            self.encryption_private == other.encryption_private
            and self.signing_public == other.signing_public
            and self.signing_private == other.signing_private
        )


class AdministrationWrapper:
    """
    Wrapper for encrypting/decrypting and signing/verifying Administration Metadata.

    Administration metadata is signed and encrypted as a JWT within a JWE using compact serialization.

    The JWT provides integrity protection via signing with an asymmetric key (allowing applications to verify metadata
    without being able to modify it). The JWE wrapper provides confidentiality via encryption with a symmetric key.

    Administration metadata is included in the JWT using a custom 'pyd' (payload) claim.

    This class checks the metadata ID corresponds to the JWT subject (internal integrity). It does not check admin
    metadata relates to discovery metadata.
    """

    _issuer = "magic.data.bas.ac.uk"
    _audience = "data.bas.ac.uk"
    _lifetime = 3_153_600_000  # 100 years
    _enc_alg = "A256GCM"

    def __init__(self, keys: AdministrationKeys) -> None:
        self._keys = keys

    def encode(self, metadata: AdministrationMetadata) -> str:
        """
        Sign and encrypt metadata.

        The JWT is signed using the private signing key (for anyone to then verify).
        The JWE is encrypted using the public encryption key (for only us to read).
        """
        if self._keys.signing_private is None:
            msg = "Private signing key is required for writing metadata."
            raise ValueError(msg) from None
        signer = JwtSigner(issuer=self._issuer, key=self._keys.signing_private, default_lifetime=self._lifetime)
        token: JweCompact = signer.sign(
            subject=metadata.id, audience=self._audience, extra_claims={"pyd": metadata.dumps_json()}
        ).encrypt(key=self._keys.encryption_private.public_jwk(), enc=self._enc_alg)
        return str(token)

    def decode(self, encrypted_metadata: str) -> AdministrationMetadata:
        """Decrypt and verify metadata."""
        token = JweCompact(encrypted_metadata)
        trusted_token = token.decrypt_jwt(self._keys.encryption_private)
        trusted_token.validate(key=self._keys.signing_public, issuer=self._issuer, audience=self._audience)
        value = AdministrationMetadata.loads_json(trusted_token.claims["pyd"])
        if trusted_token.subject != value.id:
            raise AdministrationMetadataIntegrityError() from None
        return value


def get_admin(keys: AdministrationKeys, config: dict) -> Optional[AdministrationMetadata]:
    """

    Get administration metadata for record if available.

    Checks loaded administration metadata relates to parent discovery metadata record via resource (file) identifier.
    """
    kv = get_kv(config)
    raw_value: Optional[str] = kv.get("admin_metadata", None)
    if raw_value is None:
        return None

    loader = AdministrationWrapper(keys)
    value = loader.decode(raw_value)
    if value.id != config.get("file_identifier"):
        raise AdministrationMetadataSubjectMismatchError() from None
    return value


def set_admin(keys: AdministrationKeys, config: dict, admin_meta: AdministrationMetadata) -> None:
    """Set administration metadata for record."""
    if admin_meta.id != config.get("file_identifier"):
        raise AdministrationMetadataSubjectMismatchError() from None
    wrapper = AdministrationWrapper(keys=keys)
    element = wrapper.encode(admin_meta)
    set_kv(kv={"admin_metadata": element}, config=config)


def get_kv(config: dict) -> dict:
    """Get key-value pairs from a JSON encoded string if used in record supplemental information."""
    if (
        "supplemental_information" not in config["identification"]
        or not config["identification"]["supplemental_information"]
    ):
        return {}
    try:
        kv = json.loads(config["identification"]["supplemental_information"])
    except json.JSONDecodeError:
        msg = "Supplemental information isn't JSON parsable."
        raise ValueError(msg) from None
    if not isinstance(kv, dict):
        msg = "Supplemental information isn't parsed as a dict."
        raise TypeError(msg) from None
    return kv


def set_kv(kv: dict, config: dict, replace: bool = False) -> None:
    """
    Set key-value pairs in a JSON encoded string for use in record supplemental information.

    Use `replace` to remove existing keys with new value.

    If new value is empty, removes value rather than setting an empty dict.
    """
    kv_ = get_kv(config)
    kv_.update(kv)
    if replace:
        kv_ = kv

    config["identification"]["supplemental_information"] = json.dumps(kv_)
    if len(kv_) == 0:
        del config["identification"]["supplemental_information"]



from __future__ import annotations

from lxml.etree import Element, SubElement

from bas_metadata_library import MetadataRecord
from bas_metadata_library.standards.iso_19115_common import CodeListElement, MetadataRecordElement
from bas_metadata_library.standards.iso_19115_common.common_elements import (
    AnchorElement,
    CharacterSet,
    Citation,
    Format,
    Identifier,
    Language,
    LegalConstraint,
    MaintenanceInformation,
    ResponsibleParty,
)
from bas_metadata_library.standards.iso_19115_common.utils import (
    decode_date_string,
    encode_date_string,
    format_numbers_consistently,
)


class DataIdentification(MetadataRecordElement):
    """gmd:identificationInfo."""

    def make_config(  # noqa: C901 see uk-pdc/metadata-infrastructure/metadata-library#175 for more information
        self,
    ) -> dict:
        """Decode to Python."""
        _ = {}

        citation = Citation(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:citation",
        )
        _citation = citation.make_config()
        if bool(_citation):
            _ = {**_, **_citation}

        abstract = Abstract(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:abstract",
        )
        _abstract = abstract.make_config()
        if _abstract != "":
            _["abstract"] = _abstract

        purpose = Purpose(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:purpose",
        )
        _purpose = purpose.make_config()
        if _purpose != "":
            _["purpose"] = _purpose

        credit = Credit(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:credit",
        )
        _credit = credit.make_config()
        if _credit != "":
            _["credit"] = _credit

        status = Status(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:status",
        )
        _status = status.make_config()
        if _status != "":
            _["status"] = _status

        _contacts = []
        contacts_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:pointOfContact)",
                namespaces=self.ns.nsmap(),
            )
        )
        for contact_index in range(1, contacts_length + 1):
            contact = PointOfContact(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:pointOfContact)"
                f"[{contact_index}]",
            )
            _contact = contact.make_config()
            if bool(_contact):
                _contacts.append(_contact)
        if len(_contacts) > 0:
            _["contacts"] = _contacts

        identification_maintenance = ResourceMaintenance(
            record=self.record, attributes=self.attributes, xpath=f"{self.xpath}"
        )
        _identification_maintenance = identification_maintenance.make_config()
        if bool(_identification_maintenance):
            _["maintenance"] = _identification_maintenance

        _graphic_overviews = []
        graphics_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:graphicOverview)",
                namespaces=self.ns.nsmap(),
            )
        )
        for graphic_index in range(1, graphics_length + 1):
            graphic_overview = GraphicOverview(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/"
                f"gmd:graphicOverview)[{graphic_index}]",
            )
            _graphic_overview = graphic_overview.make_config()
            if bool(_graphic_overview):
                _graphic_overviews.append(_graphic_overview)
        if len(_graphic_overviews) > 0:
            _["graphic_overviews"] = _graphic_overviews

        _resource_formats = []
        resources_formats_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:resourceFormat)",
                namespaces=self.ns.nsmap(),
            )
        )
        for resource_format_index in range(1, resources_formats_length + 1):
            resource_option_format = ResourceFormat(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:resourceFormat)"
                f"[{resource_format_index}]",
            )
            _resource_option_format = resource_option_format.make_config()
            if bool(_resource_option_format):
                _resource_formats.append(_resource_option_format)
        if len(_resource_formats) > 0:
            _["resource_formats"] = _resource_formats

        _descriptive_keywords = []
        keywords_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:descriptiveKeywords)",
                namespaces=self.ns.nsmap(),
            )
        )
        for keyword_index in range(1, keywords_length + 1):
            keywords = DescriptiveKeywords(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:descriptiveKeywords)"
                f"[{keyword_index}]",
            )
            _keywords = keywords.make_config()
            if bool(_keywords):
                _descriptive_keywords.append(_keywords)
        if len(_descriptive_keywords) > 0:
            _["keywords"] = _descriptive_keywords

        _resource_constraints = []
        constraints_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:resourceConstraints)",
                namespaces=self.ns.nsmap(),
            )
        )
        for constraint_index in range(1, constraints_length + 1):
            constraint = ResourceConstraint(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:resourceConstraints)"
                f"[{constraint_index}]",
            )
            _constraint = constraint.make_config()
            if bool(_constraint):
                _resource_constraints.append(_constraint)
        if len(_resource_constraints) > 0:
            _["constraints"] = _resource_constraints

        _resource_aggregations = []
        aggregations_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:aggregationInfo)",
                namespaces=self.ns.nsmap(),
            )
        )
        for aggregate_index in range(1, aggregations_length + 1):
            aggregation = Aggregation(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:aggregationInfo)"
                f"[{aggregate_index}]",
            )
            _aggregation = aggregation.make_config()
            if bool(_aggregation):
                _resource_aggregations.append(_aggregation)
        if len(_resource_aggregations) > 0:
            _["aggregations"] = _resource_aggregations

        spatial_representation_type = SpatialRepresentationType(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:spatialRepresentationType",
        )
        _spatial_representation_type = spatial_representation_type.make_config()
        if _spatial_representation_type != "":
            _["spatial_representation_type"] = _spatial_representation_type

        spatial_resolution = SpatialResolution(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:spatialResolution",
        )
        _spatial_resolution = spatial_resolution.make_config()
        if _spatial_resolution != "":
            _["spatial_resolution"] = _spatial_resolution

        language = Language(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:language",
        )
        _language = language.make_config()
        if _language != "":
            _["language"] = _language

        character_set = CharacterSet(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:characterSet",
        )
        _character_set = character_set.make_config()
        if _character_set != "":
            _["character_set"] = _character_set

        _topic_categories = []
        topics_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:topicCategory)",
                namespaces=self.ns.nsmap(),
            )
        )
        for topic_index in range(1, topics_length + 1):
            topic = TopicCategory(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:topicCategory)"
                f"[{topic_index}]",
            )
            _topic = topic.make_config()
            if _topic != "":
                _topic_categories.append(_topic)
        if len(_topic_categories) > 0:
            _["topics"] = _topic_categories

        _extents = []
        extents_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:extent)",
                namespaces=self.ns.nsmap(),
            )
        )
        for extent_index in range(1, extents_length + 1):
            extent = Extent(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:extent)[{extent_index}]",
            )
            _extent = extent.make_config()
            if _extent != "":
                _extents.append(_extent)
        if len(_extents) > 0:
            _["extents"] = _extents

        supplemental_information = SupplementalInformation(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:supplementalInformation",
        )
        _supplemental_information = supplemental_information.make_config()
        if _supplemental_information != "":
            _["supplemental_information"] = _supplemental_information

        return _

    def make_element(self) -> None:  # noqa: C901 see uk-pdc/metadata-infrastructure/metadata-library#175 for more information
        """Encode as XML."""
        data_identification_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}identificationInfo")
        data_identification_element = SubElement(data_identification_wrapper, f"{{{self.ns.gmd}}}MD_DataIdentification")

        citation_wrapper = SubElement(data_identification_element, f"{{{self.ns.gmd}}}citation")
        citation = Citation(
            record=self.record,
            attributes=self.attributes,
            parent_element=citation_wrapper,
            element_attributes=self.element_attributes["identification"],
        )
        citation.make_element()

        abstract = Abstract(
            record=self.record,
            attributes=self.attributes,
            parent_element=data_identification_element,
            element_attributes=self.attributes["identification"],
        )
        abstract.make_element()

        if "purpose" in self.attributes["identification"]:
            purpose = Purpose(
                record=self.record,
                attributes=self.attributes,
                parent_element=data_identification_element,
                element_attributes=self.attributes["identification"],
            )
            purpose.make_element()

        if "credit" in self.attributes["identification"]:
            credit = Credit(
                record=self.record,
                attributes=self.attributes,
                parent_element=data_identification_element,
                element_attributes=self.attributes["identification"],
            )
            credit.make_element()

        if "status" in self.attributes["identification"]:
            status = Status(
                record=self.record,
                attributes=self.attributes,
                parent_element=data_identification_element,
                element_attributes=self.attributes["identification"],
            )
            status.make_element()

        if "contacts" in self.attributes["identification"]:
            for point_of_contact_attributes in self.attributes["identification"]["contacts"]:
                for role in point_of_contact_attributes["role"]:
                    if role != "distributor":
                        _point_of_contact = point_of_contact_attributes.copy()
                        _point_of_contact["role"] = role

                        point_of_contact = PointOfContact(
                            record=self.record,
                            attributes=self.attributes,
                            parent_element=data_identification_element,
                            element_attributes=_point_of_contact,
                        )
                        point_of_contact.make_element()

        if "maintenance" in self.attributes["identification"]:
            identification_maintenance = ResourceMaintenance(
                record=self.record,
                attributes=self.attributes,
                parent_element=data_identification_element,
                element_attributes=self.element_attributes["identification"]["maintenance"],
            )
            identification_maintenance.make_element()

        if "graphic_overviews" in self.attributes["identification"]:
            for graphic_overview_attributes in self.attributes["identification"]["graphic_overviews"]:
                graphic_overview = GraphicOverview(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=data_identification_element,
                    element_attributes=graphic_overview_attributes,
                )
                graphic_overview.make_element()

        if "resource_formats" in self.attributes["identification"]:
            for resource_format_attributes in self.attributes["identification"]["resource_formats"]:
                resource_format = ResourceFormat(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=data_identification_element,
                    element_attributes=resource_format_attributes,
                )
                resource_format.make_element()

        if "keywords" in self.attributes["identification"]:
            for keyword_attributes in self.attributes["identification"]["keywords"]:
                descriptive_keywords = DescriptiveKeywords(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=data_identification_element,
                    element_attributes=keyword_attributes,
                )
                descriptive_keywords.make_element()

        if "constraints" in self.attributes["identification"]:
            for constraint_attributes in self.attributes["identification"]["constraints"]:
                constraint = ResourceConstraint(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=data_identification_element,
                    element_attributes=constraint_attributes,
                )
                constraint.make_element()

        if "aggregations" in self.attributes["identification"]:
            for aggregation_attributes in self.attributes["identification"]["aggregations"]:
                aggregation = Aggregation(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=data_identification_element,
                    element_attributes=aggregation_attributes,
                )
                aggregation.make_element()

        if "spatial_representation_type" in self.attributes["identification"]:
            spatial_representation_type = SpatialRepresentationType(
                record=self.record,
                attributes=self.attributes,
                parent_element=data_identification_element,
                element_attributes=self.attributes["identification"],
            )
            spatial_representation_type.make_element()

        if "spatial_resolution" in self.attributes["identification"]:
            spatial_resolution = SpatialResolution(
                record=self.record,
                attributes=self.attributes,
                parent_element=data_identification_element,
                element_attributes=self.attributes["identification"],
            )
            spatial_resolution.make_element()

        language = Language(
            record=self.record,
            attributes=self.attributes,
            parent_element=data_identification_element,
            element_attributes=self.attributes["identification"],
        )
        language.make_element()

        if "character_set" in self.attributes["identification"]:
            character_set = CharacterSet(
                record=self.record,
                attributes=self.attributes,
                parent_element=data_identification_element,
                element_attributes=self.attributes["identification"],
            )
            character_set.make_element()

        if "topics" in self.attributes["identification"]:
            for topic_attribute in self.attributes["identification"]["topics"]:
                topic = TopicCategory(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=data_identification_element,
                    element_attributes={"topic": topic_attribute},
                )
                topic.make_element()

        if "extents" in self.attributes["identification"]:
            for extent_attribute in self.attributes["identification"]["extents"]:
                extent = Extent(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=data_identification_element,
                    element_attributes={"extent": extent_attribute},
                )
                extent.make_element()

        if "supplemental_information" in self.attributes["identification"]:
            supplemental_information = SupplementalInformation(
                record=self.record,
                attributes=self.attributes,
                parent_element=data_identification_element,
                element_attributes=self.attributes["identification"],
            )
            supplemental_information.make_element()


class Abstract(MetadataRecordElement):
    def make_config(self) -> str:
        _ = ""

        abstract_value = self.record.xpath(f"{self.xpath}/gco:CharacterString/text()", namespaces=self.ns.nsmap())
        if len(abstract_value) == 1:
            _ = abstract_value[0]

        return _

    def make_element(self) -> None:
        abstract_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}abstract")
        abstract_value = SubElement(abstract_element, f"{{{self.ns.gco}}}CharacterString")
        abstract_value.text = self.element_attributes["abstract"]


class Purpose(MetadataRecordElement):
    def make_config(self) -> str:
        _ = ""

        purpose_value = self.record.xpath(f"{self.xpath}/gco:CharacterString/text()", namespaces=self.ns.nsmap())
        if len(purpose_value) == 1:
            _ = purpose_value[0]

        return _

    def make_element(self) -> None:
        purpose_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}purpose")
        purpose_value = SubElement(purpose_element, f"{{{self.ns.gco}}}CharacterString")
        purpose_value.text = self.element_attributes["purpose"]


class Credit(MetadataRecordElement):
    def make_config(self) -> str:
        _ = ""

        credit_value = self.record.xpath(f"{self.xpath}/gco:CharacterString/text()", namespaces=self.ns.nsmap())
        if len(credit_value) == 1:
            _ = credit_value[0]

        return _

    def make_element(self) -> None:
        credit_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}credit")
        credit_value = SubElement(credit_element, f"{{{self.ns.gco}}}CharacterString")
        credit_value.text = self.element_attributes["credit"]


class Status(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:MD_ProgressCode",
        )
        self.code_list_values = [
            "completed",
            "historicalArchive",
            "obsolete",
            "onGoing",
            "planned",
            "required",
            "underDevelopment",
        ]
        self.code_list = (
            "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/"
            "codelist/gmxCodelists.xml#MD_ProgressCode"
        )
        self.element = f"{{{self.ns.gmd}}}status"
        self.element_code = f"{{{self.ns.gmd}}}MD_ProgressCode"
        self.attribute = "status"


class PointOfContact(MetadataRecordElement):
    def make_config(self) -> dict:
        responsible_party = ResponsibleParty(record=self.record, attributes=self.attributes, xpath=self.xpath)
        return responsible_party.make_config()

    def make_element(self) -> None:
        point_of_contact_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}pointOfContact")

        responsible_party = ResponsibleParty(
            record=self.record,
            attributes=self.attributes,
            parent_element=point_of_contact_element,
            element_attributes=self.element_attributes,
        )
        responsible_party.make_element()


class ResourceMaintenance(MetadataRecordElement):
    def make_config(self) -> dict:
        maintenance_information = MaintenanceInformation(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:identificationInfo/gmd:MD_DataIdentification/gmd:resourceMaintenance",
        )
        return maintenance_information.make_config()

    def make_element(self) -> None:
        resource_maintenance_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}resourceMaintenance")
        maintenance_information = MaintenanceInformation(
            record=self.record,
            attributes=self.attributes,
            parent_element=resource_maintenance_element,
            element_attributes=self.element_attributes,
        )
        maintenance_information.make_element()


class GraphicOverview(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        identifier_value = self.record.xpath(f"{self.xpath}/gmd:MD_BrowseGraphic/@id", namespaces=self.ns.nsmap())
        if len(identifier_value) == 1:
            _["identifier"] = identifier_value[0]

        href_value = self.record.xpath(
            f"{self.xpath}/gmd:MD_BrowseGraphic/gmd:fileName/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(href_value) == 1:
            _["href"] = href_value[0]

        description_value = self.record.xpath(
            f"{self.xpath}/gmd:MD_BrowseGraphic/gmd:fileDescription/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(description_value) == 1:
            _["description"] = description_value[0]

        mime_type_value = self.record.xpath(
            f"{self.xpath}/gmd:MD_BrowseGraphic/gmd:fileType/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(mime_type_value) == 1:
            _["mime_type"] = mime_type_value[0]

        return _

    def make_element(self) -> None:
        graphic_wrapper = SubElement(
            self.parent_element,
            f"{{{self.ns.gmd}}}graphicOverview",
        )
        graphic_element = SubElement(
            graphic_wrapper,
            f"{{{self.ns.gmd}}}MD_BrowseGraphic",
            attrib={"id": self.element_attributes["identifier"]},
        )

        href_element = SubElement(graphic_element, f"{{{self.ns.gmd}}}fileName")
        href_value = SubElement(href_element, f"{{{self.ns.gco}}}CharacterString")
        href_value.text = self.element_attributes["href"]

        if "description" in self.element_attributes:
            description_element = SubElement(graphic_element, f"{{{self.ns.gmd}}}fileDescription")
            description_value = SubElement(description_element, f"{{{self.ns.gco}}}CharacterString")
            description_value.text = self.element_attributes["description"]

        if "mime_type" in self.element_attributes:
            mime_type_element = SubElement(graphic_element, f"{{{self.ns.gmd}}}fileType")
            mime_type_value = SubElement(mime_type_element, f"{{{self.ns.gco}}}CharacterString")
            mime_type_value.text = self.element_attributes["mime_type"]


class DescriptiveKeywords(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        _terms = []
        terms_values = self.record.xpath(
            f"{self.xpath}/gmd:MD_Keywords/gmd:keyword/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        for term in terms_values:
            _terms.append({"term": term})
        terms_links = self.record.xpath(
            f"{self.xpath}/gmd:MD_Keywords/gmd:keyword/gmx:Anchor", namespaces=self.ns.nsmap()
        )
        for term in terms_links:
            _term = {}
            _term_value = term.xpath("./text()", namespaces=self.ns.nsmap())
            if len(_term_value) > 0:
                _term["term"] = _term_value[0]
            _term_href = term.xpath("./@xlink:href", namespaces=self.ns.nsmap())
            if len(_term_href) > 0:
                _term["href"] = _term_href[0]
            if bool(_term):
                _terms.append(_term)
        if bool(_terms):
            _["terms"] = _terms

        descriptive_keywords_type = DescriptiveKeywordsType(
            record=self.record, attributes=self.attributes, xpath=f"{self.xpath}/gmd:MD_Keywords/gmd:type"
        )
        _descriptive_keywords_type = descriptive_keywords_type.make_config()
        if _descriptive_keywords_type != "":
            _["type"] = _descriptive_keywords_type

        thesaurus = Thesaurus(
            record=self.record, attributes=self.attributes, xpath=f"{self.xpath}/gmd:MD_Keywords/gmd:thesaurusName"
        )
        _thesaurus = thesaurus.make_config()
        if bool(_thesaurus):
            _["thesaurus"] = _thesaurus

        return _

    def make_element(self) -> None:
        keywords_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}descriptiveKeywords")
        keywords_element = SubElement(keywords_wrapper, f"{{{self.ns.gmd}}}MD_Keywords")

        for term in self.element_attributes["terms"]:
            term_element = SubElement(keywords_element, f"{{{self.ns.gmd}}}keyword")
            if "href" in term:
                anchor = AnchorElement(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=term_element,
                    element_attributes=term,
                    element_value=term["term"],
                )
                anchor.make_element()
            else:
                term_value = SubElement(term_element, f"{{{self.ns.gco}}}CharacterString")
                term_value.text = term["term"]

        if "type" in self.element_attributes:
            keyword_type = DescriptiveKeywordsType(
                record=self.record,
                attributes=self.attributes,
                parent_element=keywords_element,
                element_attributes=self.element_attributes,
            )
            keyword_type.make_element()

        if "thesaurus" in self.element_attributes:
            thesaurus = Thesaurus(
                record=self.record,
                attributes=self.attributes,
                parent_element=keywords_element,
                element_attributes=self.element_attributes["thesaurus"],
            )
            thesaurus.make_element()


class DescriptiveKeywordsType(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:MD_KeywordTypeCode",
        )
        self.code_list_values = ["discipline", "place", "stratum", "temporal", "theme"]
        self.code_list = (
            "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/"
            "codelist/gmxCodelists.xml#MD_KeywordTypeCode"
        )
        self.element = f"{{{self.ns.gmd}}}type"
        self.element_code = f"{{{self.ns.gmd}}}MD_KeywordTypeCode"
        self.attribute = "type"


class Thesaurus(MetadataRecordElement):
    def make_config(self) -> dict:
        citation = Citation(record=self.record, attributes=self.attributes, xpath=self.xpath)
        return citation.make_config()

    def make_element(self) -> None:
        thesaurus_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}thesaurusName")

        citation = Citation(
            record=self.record,
            attributes=self.attributes,
            parent_element=thesaurus_element,
            element_attributes=self.element_attributes,
        )
        citation.make_element()


class ResourceConstraint(MetadataRecordElement):
    def make_config(self) -> dict:
        legal_constraint = LegalConstraint(
            record=self.record,
            attributes=self.attributes,
            xpath=self.xpath,
        )
        return legal_constraint.make_config()

    def make_element(self) -> None:
        constraints_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}resourceConstraints")

        legal_constraint = LegalConstraint(
            record=self.record,
            attributes=self.attributes,
            parent_element=constraints_wrapper,
            element_attributes=self.element_attributes,
        )
        legal_constraint.make_element()


class Aggregation(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        association_type = AssociationType(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:MD_AggregateInformation/gmd:associationType",
        )
        _association_type = association_type.make_config()
        if _association_type != "":
            _["association_type"] = _association_type

        initiative_type = InitiativeType(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:MD_AggregateInformation/gmd:initiativeType",
        )
        _initiative_type = initiative_type.make_config()
        if _initiative_type != "":
            _["initiative_type"] = _initiative_type

        identifier = Identifier(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:MD_AggregateInformation/gmd:aggregateDataSetIdentifier",
        )
        _identifier = identifier.make_config()
        if bool(_identifier):
            _["identifier"] = _identifier

        return _

    def make_element(self) -> None:
        aggregation_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}aggregationInfo")
        aggregation_element = SubElement(aggregation_wrapper, f"{{{self.ns.gmd}}}MD_AggregateInformation")

        identifier = Identifier(
            record=self.record,
            attributes=self.attributes,
            parent_element=aggregation_element,
            element_attributes=self.element_attributes["identifier"],
            identifier_container=f"{{{self.ns.gmd}}}aggregateDataSetIdentifier",
        )
        identifier.make_element()

        association_type = AssociationType(
            record=self.record,
            attributes=self.attributes,
            parent_element=aggregation_element,
            element_attributes=self.element_attributes,
        )
        association_type.make_element()

        if "initiative_type" in self.element_attributes:
            initiative_type = InitiativeType(
                record=self.record,
                attributes=self.attributes,
                parent_element=aggregation_element,
                element_attributes=self.element_attributes,
            )
            initiative_type.make_element()


class AssociationType(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:DS_AssociationTypeCode",
        )
        self.code_list_values = [
            "collectiveTitle",
            "crossReference",
            "dependency",
            "isComposedOf",
            "largerWorkCitation",
            "partOfSeamlessDatabase",
            "revisionOf",
            "series",
            "stereoMate",
            "physicalReverseOf",
        ]
        self.code_list = (
            "https://standards.iso.org/iso/19115/resources/Codelists/cat/codelists.xml#DS_AssociationTypeCode"
        )
        self.element = f"{{{self.ns.gmd}}}associationType"
        self.element_code = f"{{{self.ns.gmd}}}DS_AssociationTypeCode"
        self.attribute = "association_type"


class InitiativeType(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:DS_InitiativeTypeCode",
        )
        self.code_list_values = [
            "campaign",
            "collection",
            "exercise",
            "experiment",
            "investigation",
            "mission",
            "operation",
            "platform",
            "process",
            "program",
            "project",
            "sensor",
            "study",
            "task",
            "trial",
            "dataDictionary",
            "sciencePaper",
            "userGuide",
        ]
        self.code_list = (
            "https://standards.iso.org/iso/19115/resources/Codelists/cat/codelists.xml#DS_InitiativeTypeCode"
        )
        self.element = f"{{{self.ns.gmd}}}initiativeType"
        self.element_code = f"{{{self.ns.gmd}}}DS_InitiativeTypeCode"
        self.attribute = "initiative_type"


class SupplementalInformation(MetadataRecordElement):
    def make_config(self) -> str:
        _ = ""

        supplemental_value = self.record.xpath(f"{self.xpath}/gco:CharacterString/text()", namespaces=self.ns.nsmap())
        if len(supplemental_value) == 1:
            _ = supplemental_value[0]

        return _

    def make_element(self) -> None:
        if "supplemental_information" in self.element_attributes:
            supplemental_info_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}supplementalInformation")
            supplemental_info_value = SubElement(supplemental_info_element, f"{{{self.ns.gco}}}CharacterString")
            supplemental_info_value.text = self.element_attributes["supplemental_information"]


class SpatialRepresentationType(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:MD_SpatialRepresentationTypeCode",
        )
        self.code_list_values = ["vector", "grid", "textTable", "tin", "stereoModel", "video"]
        self.code_list = (
            "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/"
            "codelist/gmxCodelists.xml#MD_SpatialRepresentationTypeCode"
        )
        self.element = f"{{{self.ns.gmd}}}spatialRepresentationType"
        self.element_code = f"{{{self.ns.gmd}}}MD_SpatialRepresentationTypeCode"
        self.attribute = "spatial_representation_type"


class SpatialResolution(MetadataRecordElement):
    def make_config(self) -> str:
        _ = ""

        spatial_resolution_value = self.record.xpath(
            f"{self.xpath}/gmd:MD_Resolution/gmd:equivalentScale/gmd:MD_RepresentativeFraction/gmd:denominator/"
            f"gco:Integer/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(spatial_resolution_value) == 1:
            _ = format_numbers_consistently(spatial_resolution_value[0])

        return _

    def make_element(self) -> None:
        if self.element_attributes["spatial_resolution"] is None:
            SubElement(
                self.parent_element,
                f"{{{self.ns.gmd}}}spatialResolution",
                attrib={f"{{{self.ns.gco}}}nilReason": "inapplicable"},
            )
        else:
            resolution_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}spatialResolution")
            resolution_element = SubElement(resolution_wrapper, f"{{{self.ns.gmd}}}MD_Resolution")
            equivalent_scale_wrapper = SubElement(resolution_element, f"{{{self.ns.gmd}}}equivalentScale")
            equivalent_scale_element = SubElement(
                equivalent_scale_wrapper, f"{{{self.ns.gmd}}}MD_RepresentativeFraction"
            )
            denominator_element = SubElement(equivalent_scale_element, f"{{{self.ns.gmd}}}denominator")
            denominator_value = SubElement(denominator_element, f"{{{self.ns.gco}}}Integer")
            denominator_value.text = str(self.element_attributes["spatial_resolution"])


class TopicCategory(MetadataRecordElement):
    def make_config(self) -> str:
        _ = ""

        topic_category_value = self.record.xpath(
            f"{self.xpath}/gmd:MD_TopicCategoryCode/text()", namespaces=self.ns.nsmap()
        )
        if len(topic_category_value) == 1:
            _ = topic_category_value[0]

        return _

    def make_element(self) -> None:
        topic_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}topicCategory")
        topic_value = SubElement(topic_element, f"{{{self.ns.gmd}}}MD_TopicCategoryCode")
        topic_value.text = self.element_attributes["topic"]


class Extent(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        identifier_value = self.record.xpath(
            f"{self.xpath}/gmd:EX_Extent/@id",
            namespaces=self.ns.nsmap(),
        )
        if len(identifier_value) == 1:
            _["identifier"] = identifier_value[0]

        geographic_extent = GeographicExtent(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:EX_Extent",
        )
        _geographic_extent = geographic_extent.make_config()
        if bool(_geographic_extent):
            _["geographic"] = _geographic_extent

        temporal_extent = TemporalExtent(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:EX_Extent",
        )
        _temporal_extent = temporal_extent.make_config()
        if bool(_temporal_extent):
            _["temporal"] = _temporal_extent

        vertical_extent = VerticalExtent(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:EX_Extent/gmd:verticalElement",
        )
        _vertical_extent = vertical_extent.make_config()
        if bool(_vertical_extent):
            _["vertical"] = _vertical_extent

        return _

    def make_element(self) -> None:
        extent_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}extent")
        extent_element = SubElement(
            extent_wrapper,
            f"{{{self.ns.gmd}}}EX_Extent",
            attrib={"id": self.element_attributes["extent"]["identifier"]},
        )

        for extent_item in self.element_attributes.values():
            if "geographic" in extent_item:
                geographic_extent = GeographicExtent(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=extent_element,
                    element_attributes=extent_item["geographic"],
                )
                geographic_extent.make_element()

            if "temporal" in extent_item:
                temporal_extent = TemporalExtent(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=extent_element,
                    element_attributes=extent_item["temporal"],
                )
                temporal_extent.make_element()

            if "vertical" in extent_item:
                vertical_extent = VerticalExtent(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=extent_element,
                    element_attributes=extent_item["vertical"],
                )
                vertical_extent.make_element()


class GeographicExtent(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        bounding_box = BoundingBox(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:geographicElement",
        )
        _bounding_box = bounding_box.make_config()
        if bool(_bounding_box):
            _["bounding_box"] = _bounding_box

        identifier = Identifier(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:geographicElement/gmd:EX_GeographicDescription/gmd:geographicIdentifier",
        )
        _identifier = identifier.make_config()
        if bool(_identifier):
            _["identifier"] = _identifier

        return _

    def make_element(self) -> None:
        geographic_extent_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}geographicElement")

        if "bounding_box" in self.element_attributes:
            bounding_box = BoundingBox(
                record=self.record,
                attributes=self.attributes,
                parent_element=geographic_extent_element,
                element_attributes=self.element_attributes["bounding_box"],
            )
            bounding_box.make_element()
        elif "identifier" in self.element_attributes:
            geographic_description_element = SubElement(
                geographic_extent_element, f"{{{self.ns.gmd}}}EX_GeographicDescription"
            )
            identifier = Identifier(
                record=self.record,
                attributes=self.attributes,
                parent_element=geographic_description_element,
                element_attributes=self.element_attributes["identifier"],
            )
            identifier.identifier_container = f"{{{self.ns.gmd}}}geographicIdentifier"
            identifier.make_element()


class BoundingBox(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        west_element_value = self.record.xpath(
            f"{self.xpath}/gmd:EX_GeographicBoundingBox/gmd:westBoundLongitude/gco:Decimal/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(west_element_value) == 1:
            _["west_longitude"] = float(west_element_value[0])

        east_element_value = self.record.xpath(
            f"{self.xpath}/gmd:EX_GeographicBoundingBox/gmd:eastBoundLongitude/gco:Decimal/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(east_element_value) == 1:
            _["east_longitude"] = float(east_element_value[0])

        south_element_value = self.record.xpath(
            f"{self.xpath}/gmd:EX_GeographicBoundingBox/gmd:southBoundLatitude/gco:Decimal/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(south_element_value) == 1:
            _["south_latitude"] = float(south_element_value[0])

        north_element_value = self.record.xpath(
            f"{self.xpath}/gmd:EX_GeographicBoundingBox/gmd:northBoundLatitude/gco:Decimal/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(north_element_value) == 1:
            _["north_latitude"] = float(north_element_value[0])

        return _

    def make_element(self) -> None:
        bounding_box_element = SubElement(
            self.parent_element,
            f"{{{self.ns.gmd}}}EX_GeographicBoundingBox",
        )

        west_element = SubElement(bounding_box_element, f"{{{self.ns.gmd}}}westBoundLongitude")
        west_value = SubElement(west_element, f"{{{self.ns.gco}}}Decimal")
        west_value.text = str(self.element_attributes["west_longitude"])

        east_element = SubElement(bounding_box_element, f"{{{self.ns.gmd}}}eastBoundLongitude")
        east_value = SubElement(east_element, f"{{{self.ns.gco}}}Decimal")
        east_value.text = str(self.element_attributes["east_longitude"])

        south_element = SubElement(bounding_box_element, f"{{{self.ns.gmd}}}southBoundLatitude")
        south_value = SubElement(south_element, f"{{{self.ns.gco}}}Decimal")
        south_value.text = str(self.element_attributes["south_latitude"])

        north_element = SubElement(bounding_box_element, f"{{{self.ns.gmd}}}northBoundLatitude")
        north_value = SubElement(north_element, f"{{{self.ns.gco}}}Decimal")
        north_value.text = str(self.element_attributes["north_latitude"])


class VerticalExtent(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        minimum_value = self.record.xpath(
            f"{self.xpath}/gmd:EX_VerticalExtent/gmd:minimumValue/gco:Real/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(minimum_value) == 1:
            _["minimum"] = float(minimum_value[0])

        maximum_value = self.record.xpath(
            f"{self.xpath}/gmd:EX_VerticalExtent/gmd:maximumValue/gco:Real/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(maximum_value) == 1:
            _["maximum"] = float(maximum_value[0])

        vertical_crs = VerticalCRS(
            record=self.record, attributes=self.attributes, xpath=f"{self.xpath}/gmd:EX_VerticalExtent"
        )
        _vertical_crs = vertical_crs.make_config()
        if bool(_vertical_crs):
            _ = {**_, **_vertical_crs}

        return _

    def make_element(self) -> None:
        vertical_extent_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}verticalElement")
        vertical_extent_element = SubElement(vertical_extent_wrapper, f"{{{self.ns.gmd}}}EX_VerticalExtent")

        if "minimum" in self.element_attributes:
            minimum_element = SubElement(vertical_extent_element, f"{{{self.ns.gmd}}}minimumValue")
            minimum_value = SubElement(minimum_element, f"{{{self.ns.gco}}}Real")
            minimum_value.text = str(self.element_attributes["minimum"])

        if "maximum" in self.element_attributes:
            maximum_element = SubElement(vertical_extent_element, f"{{{self.ns.gmd}}}maximumValue")
            maximum_value = SubElement(maximum_element, f"{{{self.ns.gco}}}Real")
            maximum_value.text = str(self.element_attributes["maximum"])

        if "code" in self.element_attributes:
            vertical_crs = VerticalCRS(
                record=self.record,
                attributes=self.attributes,
                parent_element=vertical_extent_element,
                element_attributes=self.element_attributes,
            )
            vertical_crs.make_element()


class VerticalCRS(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        identifier_value = self.record.xpath(
            f"{self.xpath}/gmd:verticalCRS/gml:VerticalCRS/@gml:id",
            namespaces=self.ns.nsmap(),
        )
        if len(identifier_value) == 1:
            _["identifier"] = identifier_value[0]

        code_value = self.record.xpath(
            f"{self.xpath}/gmd:verticalCRS/gml:VerticalCRS/gml:identifier/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(code_value) == 1:
            _["code"] = code_value[0]

        name_value = self.record.xpath(
            f"{self.xpath}/gmd:verticalCRS/gml:VerticalCRS/gml:name/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(name_value) == 1:
            _["name"] = name_value[0]

        remarks_value = self.record.xpath(
            f"{self.xpath}/gmd:verticalCRS/gml:VerticalCRS/gml:remarks/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(remarks_value) == 1:
            _["remarks"] = remarks_value[0]

        domain_of_validity_href = self.record.xpath(
            f"{self.xpath}/gmd:verticalCRS/gml:VerticalCRS/gml:domainOfValidity/@xlink:href",
            namespaces=self.ns.nsmap(),
        )
        if len(domain_of_validity_href) == 1:
            _["domain_of_validity"] = {"href": domain_of_validity_href[0]}

        scope_value = self.record.xpath(
            f"{self.xpath}/gmd:verticalCRS/gml:VerticalCRS/gml:scope/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(scope_value) == 1:
            _["scope"] = scope_value[0]

        vertical_cs_href = self.record.xpath(
            f"{self.xpath}/gmd:verticalCRS/gml:VerticalCRS/gml:verticalCS/@xlink:href",
            namespaces=self.ns.nsmap(),
        )
        if len(vertical_cs_href) == 1:
            _["vertical_cs"] = {"href": vertical_cs_href[0]}

        vertical_datum_href = self.record.xpath(
            f"{self.xpath}/gmd:verticalCRS/gml:VerticalCRS/gml:verticalDatum/@xlink:href",
            namespaces=self.ns.nsmap(),
        )
        if len(vertical_datum_href) == 1:
            _["vertical_datum"] = {"href": vertical_datum_href[0]}

        return _

    def make_element(self) -> None:
        vertical_crs_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}verticalCRS")
        vertical_crs_element = SubElement(
            vertical_crs_wrapper,
            f"{{{self.ns.gml}}}VerticalCRS",
            attrib={f"{{{self.ns.gml}}}id": self.element_attributes["identifier"]},
        )
        vertical_crs_code = SubElement(
            vertical_crs_element, f"{{{self.ns.gml}}}identifier", attrib={"codeSpace": "OGP"}
        )
        vertical_crs_code.text = self.element_attributes["code"]

        name = SubElement(vertical_crs_element, f"{{{self.ns.gml}}}name")
        name.text = self.element_attributes["name"]

        remarks = SubElement(vertical_crs_element, f"{{{self.ns.gml}}}remarks")
        remarks.text = self.element_attributes["remarks"]

        SubElement(
            vertical_crs_element,
            f"{{{self.ns.gml}}}domainOfValidity",
            attrib={f"{{{self.ns.xlink}}}href": self.element_attributes["domain_of_validity"]["href"]},
        )

        scope = SubElement(vertical_crs_element, f"{{{self.ns.gml}}}scope")
        scope.text = self.element_attributes["scope"]

        SubElement(
            vertical_crs_element,
            f"{{{self.ns.gml}}}verticalCS",
            attrib={f"{{{self.ns.xlink}}}href": self.element_attributes["vertical_cs"]["href"]},
        )

        SubElement(
            vertical_crs_element,
            f"{{{self.ns.gml}}}verticalDatum",
            attrib={f"{{{self.ns.xlink}}}href": self.element_attributes["vertical_datum"]["href"]},
        )


class TemporalExtent(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        begin_value = self.record.xpath(
            f"{self.xpath}/gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml:TimePeriod/gml:beginPosition/"
            f"text()",
            namespaces=self.ns.nsmap(),
        )
        if len(begin_value) == 1:
            if "period" not in _:
                _["period"] = {}
            try:
                _["period"]["start"] = decode_date_string(date_datetime=begin_value[0])
            except ValueError:
                msg = "Date/datetime could not be parsed as an ISO date value"
                raise RuntimeError(msg) from None

        end_value = self.record.xpath(
            f"{self.xpath}/gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml:TimePeriod/gml:endPosition/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(end_value) == 1:
            if "period" not in _:
                _["period"] = {}
            try:
                _["period"]["end"] = decode_date_string(date_datetime=end_value[0])
            except ValueError:
                msg = "Date/datetime could not be parsed as an ISO date value"
                raise RuntimeError(msg) from None

        return _

    def make_element(self) -> None:
        temporal_extent_container = SubElement(self.parent_element, f"{{{self.ns.gmd}}}temporalElement")
        temporal_extent_wrapper = SubElement(temporal_extent_container, f"{{{self.ns.gmd}}}EX_TemporalExtent")
        temporal_extent_element = SubElement(temporal_extent_wrapper, f"{{{self.ns.gmd}}}extent")

        if "period" in self.element_attributes:
            time_period_element = SubElement(
                temporal_extent_element,
                f"{{{self.ns.gml}}}TimePeriod",
            )

            _date_precision = None
            if "date_precision" in self.element_attributes["period"]["start"]:
                _date_precision = self.element_attributes["period"]["start"]["date_precision"]
            begin_position_element = SubElement(time_period_element, f"{{{self.ns.gml}}}beginPosition")
            begin_position_element.text = encode_date_string(
                date_datetime=self.element_attributes["period"]["start"]["date"],
                date_precision=_date_precision,
            )

            end_position_element = SubElement(time_period_element, f"{{{self.ns.gml}}}endPosition")
            if "end" in self.element_attributes["period"]:
                _date_precision = None
                if "date_precision" in self.element_attributes["period"]["end"]:
                    _date_precision = self.element_attributes["period"]["end"]["date_precision"]
                end_position_element.text = encode_date_string(
                    date_datetime=self.element_attributes["period"]["end"]["date"],
                    date_precision=_date_precision,
                )
            else:
                end_position_element.attrib["indeterminatePosition"] = "unknown"


class ResourceFormat(MetadataRecordElement):
    def make_config(self) -> dict:
        resource_format = Format(record=self.record, attributes=self.attributes, xpath=self.xpath)
        return resource_format.make_config()

    def make_element(self) -> None:
        resource_format_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}resourceFormat")

        resource_format = Format(
            record=self.record,
            attributes=self.attributes,
            parent_element=resource_format_element,
            element_attributes=self.element_attributes,
        )
        resource_format.make_element()



from __future__ import annotations

from lxml.etree import Element, SubElement

from bas_metadata_library import (
    MetadataRecord as _MetadataRecord,
)
from bas_metadata_library import (
    MetadataRecordElement as _MetadataRecordElement,
)
from bas_metadata_library import (
    Namespaces as _Namespaces,
)


class Namespaces(_Namespaces):
    """Defines the namespaces for this standard."""

    gmd = "http://www.isotc211.org/2005/gmd"
    gco = "http://www.isotc211.org/2005/gco"
    gml = "http://www.opengis.net/gml/3.2"
    gmx = "http://www.isotc211.org/2005/gmx"
    srv = "http://www.isotc211.org/2005/srv"
    xlink = "http://www.w3.org/1999/xlink"
    xsi = "http://www.w3.org/2001/XMLSchema-instance"
    gmi = "http://www.isotc211.org/2005/gmi"
    gss = "http://www.isotc211.org/2005/gss"
    gsr = "http://www.isotc211.org/2005/gsr"
    gts = "http://www.isotc211.org/2005/gts"

    _schema_locations = {  # noqa: RUF012
        "gmd": "https://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/gmd/gmd.xsd",
        "gco": "https://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/gco/gco.xsd",
        "gmx": "https://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/gmx/gmx.xsd",
        "srv": "https://standards.iso.org/iso/19119/srv/srv.xsd",
        "gmi": "https://standards.iso.org/iso/19115/-2/gmi/1.0/gmi.xsd",
        "gss": "https://standards.iso.org/iso/19139/Schemas/gss/gss.xsd",
        "gsr": "https://standards.iso.org/iso/19139/Schemas/gsr/gsr.xsd",
        "gts": "https://standards.iso.org/iso/19139/Schemas/gts/gts.xsd",
    }

    def __init__(self):
        self._namespaces = {
            "gmd": self.gmd,
            "gco": self.gco,
            "gml": self.gml,
            "gmx": self.gmx,
            "srv": self.srv,
            "xlink": self.xlink,
            "xsi": self.xsi,
            "gmi": self.gmi,
            "gss": self.gss,
            "gsr": self.gsr,
            "gts": self.gts,
        }
        super().__init__(namespaces=self._namespaces)


class MetadataRecordElement(_MetadataRecordElement):
    """
    Overloaded base MetadataRecordElement class.

    Sets the type hint of the record attribute to the MetadataRecord class for this metadata standard.
    """

    def __init__(
        self,
        record: _MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=xpath,
        )
        self.ns = Namespaces()


class CodeListElement(MetadataRecordElement):
    """Derived MetadataRecordElement class defining an ISO code list element."""

    def __init__(
        self,
        record: _MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=xpath,
        )
        self.code_list_values = []
        self.code_list = None
        self.element = None
        self.element_code = None
        self.attribute = None

    def make_config(self) -> str:
        """Build partial record configuration."""
        _ = ""

        value = self.record.xpath(
            f"{self.xpath}[@codeList = '{self.code_list}']/@codeListValue",
            namespaces=self.ns.nsmap(),
        )
        if len(value) == 1:
            _ = value[0]

        return _

    def make_element(self) -> None:
        """Build XML element."""
        code_list_element = SubElement(self.parent_element, self.element)
        if (
            self.attribute in self.element_attributes
            and self.element_attributes[self.attribute] in self.code_list_values
        ):
            code_list_value = SubElement(
                code_list_element,
                self.element_code,
                attrib={"codeList": self.code_list, "codeListValue": self.element_attributes[self.attribute]},
            )
            code_list_value.text = self.element_attributes[self.attribute]



import json
from dataclasses import asdict, dataclass, field
from datetime import datetime, timezone
from typing import Final, Optional, TypeVar

import cattrs

TAdministrationMetadata = TypeVar("TAdministration", bound="AdministrationMetadata")


@dataclass
class Permission:
    """
    Access permission.

    Represents a group with access to a resource or resource description within a given directory for a given time.

    Where not specified, the expiry is set to the maximum future time by default representing an unbounded duration.

    A freetext comment can optionally be added for any additional context (e.g. why a non-obvious group has access).

    Note: Comments are ignored in equality comparisons.

    Schema definition: permission [1]

    [1] https://github.com/antarctica/metadata-library/blob/v0.16.0/src/bas_metadata_library/schemas/dist/magic_administration_content_v1.json#L7
    """

    directory: str
    group: str
    expiry: datetime = field(default_factory=lambda: datetime.max.replace(tzinfo=timezone.utc))
    comment: Optional[str] = None

    def __eq__(self, other: object) -> bool:
        """Equality comparison ignoring comment."""
        if not isinstance(other, type(self)):
            raise TypeError() from None
        self_dict = asdict(self)
        other_dict = asdict(other)
        self_dict.pop("comment", None)
        other_dict.pop("comment", None)
        return self_dict == other_dict


@dataclass
class AdministrationMetadata:
    """
    Representation of non-public, non-standard, metadata for internal administrative use.

    This class is a complete mapping of the BAS MAGIC Administration Metadata v1 content schema [1] to Python
    dataclasses.

    This class supports conversion to/from plain types and JSON strings. It does not support signing and/or encryption.

    Schema definition: administration [2]

    Note: in Python 3.10+ set kw_only=True in dataclass decorator.

    [1] https://github.com/antarctica/metadata-library/blob/v0.16.0/src/bas_metadata_library/schemas/dist/magic_administration_content_v1.json
    [2] https://github.com/antarctica/metadata-library/blob/v0.16.0/src/bas_metadata_library/schemas/dist/magic_administration_content_v1.json-L107
    """

    id: str
    gitlab_issues: list[str] = field(default_factory=list)
    metadata_permissions: list[Permission] = field(default_factory=list)
    resource_permissions: list[Permission] = field(default_factory=list)
    _schema: Final[str] = (
        "https://metadata-resources.data.bas.ac.uk/bas-metadata-generator-configuration-schemas/v2/magic-administration-content-v1.json"
    )

    def __post_init__(self) -> None:
        """Validate."""
        if not self.id:
            msg = "Administration metadata requires an id."
            raise TypeError(msg) from None
        for issue in self.gitlab_issues:
            if "/-/issues/" not in issue:
                msg = f"URL '{issue}' is not a valid GitLab issue."
                raise ValueError(msg) from None

    @classmethod
    def structure(cls: type[TAdministrationMetadata], value: dict) -> "AdministrationMetadata":
        """
        Parse AdministrationMetadata class from plain types.

        Returns a new class instance with parsed data. Intended to be used as a cattrs structure hook.
        E.g. `converter.register_structure_hook(AdministrationMetadata, lambda d, t: AdministrationMetadata.structure(d))`
        """
        if value["$schema"] != AdministrationMetadata._schema:
            msg = "Unsupported JSON Schema in data."
            raise ValueError(msg) from None

        converter = cattrs.Converter()
        converter.register_structure_hook(datetime, lambda d, t: datetime.fromisoformat(d))
        return converter.structure(value, cls)

    def unstructure(self) -> dict:
        """
        Convert AdministrationMetadata class into plain types.

        Intended to be used as a cattrs unstructure hook.
        E.g. `converter.register_unstructure_hook(AdministrationMetadata, lambda d: d.unstructure())`
        """
        converter = cattrs.Converter()
        converter.register_unstructure_hook(datetime, lambda d: d.isoformat())
        value = converter.unstructure(self)

        # remove internal keys (ensuring order)
        value.pop("_schema", None)

        return value

    @classmethod
    def loads_json(cls: type[TAdministrationMetadata], value: str) -> "AdministrationMetadata":
        """Parse AdministrationMetadata class from a JSON encoded string."""
        return AdministrationMetadata.structure(json.loads(value))

    def dumps_json(self) -> str:
        """Convert AdministrationMetadata class into a JSON encoded string."""
        return json.dumps({"$schema": self._schema, **self.unstructure()}, indent=2, ensure_ascii=False)



from __future__ import annotations

from datetime import date

from lxml.etree import Element, SubElement

from bas_metadata_library import MetadataRecord
from bas_metadata_library.standards.iso_19115_common import CodeListElement, MetadataRecordElement
from bas_metadata_library.standards.iso_19115_common.common_elements import (
    AnchorElement,
    Citation,
    LegalConstraint,
    MaintenanceInformation,
    ResponsibleParty,
)
from bas_metadata_library.standards.iso_19115_common.utils import decode_date_string, encode_date_string


class FileIdentifier(MetadataRecordElement):
    def make_config(self) -> str:
        _ = ""

        value = self.record.xpath(
            f"{self.xpath}/gmd:fileIdentifier/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(value) == 1:
            _ = value[0]

        return _

    def make_element(self) -> None:
        if "file_identifier" in self.attributes:
            file_identifier_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}fileIdentifier")
            file_identifier_value = SubElement(file_identifier_element, f"{{{self.ns.gco}}}CharacterString")
            file_identifier_value.text = self.attributes["file_identifier"]


class ScopeCode(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:MD_ScopeCode",
        )
        self.code_list_values = [
            "aggregate",
            "application",
            "attribute",
            "attributeType",
            "collection",
            "collectionHardware",
            "collectionSession",
            "coverage",
            "dataset",
            "dimensionGroup",
            "document",
            "feature",
            "featureType",
            "fieldSession",
            "initiative",
            "metadata",
            "model",
            "nonGeographicDataset",
            "product",
            "propertyType",
            "repository",
            "sample",
            "series",
            "service",
            "software",
            "tile",
        ]
        self.code_list = "https://standards.iso.org/iso/19115/resources/Codelists/cat/codelists.xml#MD_ScopeCode"
        self.element = f"{{{self.ns.gmd}}}level"
        self.element_code = f"{{{self.ns.gmd}}}MD_ScopeCode"
        self.attribute = "hierarchy_level"


class HierarchyLevel(ScopeCode):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=xpath,
        )
        self.element = f"{{{self.ns.gmd}}}hierarchyLevel"

    def make_element(self) -> None:
        super().make_element()
        hierarchy_level_name_element = SubElement(self.record, f"{{{self.ns.gmd}}}hierarchyLevelName")
        if self.attribute in self.attributes and self.attributes[self.attribute] in self.code_list_values:
            hierarchy_level_name_value = SubElement(hierarchy_level_name_element, f"{{{self.ns.gco}}}CharacterString")
            hierarchy_level_name_value.text = self.attributes[self.attribute]


class Contact(MetadataRecordElement):
    @property
    def make_config(self) -> dict:
        responsible_party = ResponsibleParty(record=self.record, attributes=self.attributes, xpath=self.xpath)
        return responsible_party.make_config()

    def make_element(self) -> None:
        contact_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}contact")

        responsible_party = ResponsibleParty(
            record=self.record,
            attributes=self.attributes,
            parent_element=contact_element,
            element_attributes=self.element_attributes,
        )
        responsible_party.make_element()


class DateStamp(MetadataRecordElement):
    def make_config(self) -> date | None:
        _ = None

        value = self.record.xpath(f"{self.xpath}/gmd:dateStamp/gco:Date/text()", namespaces=self.ns.nsmap())
        if len(value) == 1:
            try:
                _ = decode_date_string(date_datetime=value[0])["date"]
            except ValueError:
                msg = "Datestamp could not be parsed as an ISO date value"
                raise RuntimeError(msg) from None

        return _

    def make_element(self) -> None:
        date_stamp_element = SubElement(self.record, f"{{{self.ns.gmd}}}dateStamp")
        date_stamp_value = SubElement(date_stamp_element, f"{{{self.ns.gco}}}Date")
        date_stamp_value.text = encode_date_string(self.attributes["date_stamp"])


class MetadataMaintenance(MetadataRecordElement):
    def make_config(self) -> dict:
        maintenance_information = MaintenanceInformation(
            record=self.record, attributes=self.attributes, xpath=f"{self.xpath}/gmd:metadataMaintenance"
        )
        return maintenance_information.make_config()

    def make_element(self) -> None:
        if "maintenance" in self.attributes["metadata"]:
            metadata_maintenance_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}metadataMaintenance")

            maintenance_information = MaintenanceInformation(
                record=self.record,
                attributes=self.attributes,
                parent_element=metadata_maintenance_element,
                element_attributes=self.attributes["metadata"]["maintenance"],
            )
            maintenance_information.make_element()


class MetadataConstraint(MetadataRecordElement):
    def make_config(self) -> dict:
        legal_constraint = LegalConstraint(
            record=self.record,
            attributes=self.attributes,
            xpath=self.xpath,
        )
        return legal_constraint.make_config()

    def make_element(self) -> None:
        constraints_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}metadataConstraints")

        legal_constraint = LegalConstraint(
            record=self.record,
            attributes=self.attributes,
            parent_element=constraints_wrapper,
            element_attributes=self.element_attributes,
        )
        legal_constraint.make_element()


class MetadataStandard(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        standard_name = self.record.xpath(
            f"/{self.xpath}/gmd:metadataStandardName/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(standard_name) == 1:
            _["name"] = standard_name[0]

        standard_version = self.record.xpath(
            f"/{self.xpath}/gmd:metadataStandardVersion/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(standard_version) == 1:
            _["version"] = standard_version[0]

        return _

    def make_element(self) -> None:
        if "name" in self.element_attributes:
            metadata_standard_name_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}metadataStandardName")
            metadata_standard_name_value = SubElement(
                metadata_standard_name_element, f"{{{self.ns.gco}}}CharacterString"
            )
            metadata_standard_name_value.text = self.element_attributes["name"]

        if "version" in self.element_attributes:
            metadata_standard_version_element = SubElement(
                self.parent_element, f"{{{self.ns.gmd}}}metadataStandardVersion"
            )
            metadata_standard_version_value = SubElement(
                metadata_standard_version_element, f"{{{self.ns.gco}}}CharacterString"
            )
            metadata_standard_version_value.text = self.element_attributes["version"]


class ReferenceSystemInfo(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        authority = Citation(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:referenceSystemInfo/gmd:MD_ReferenceSystem/gmd:referenceSystemIdentifier/"
            f"gmd:RS_Identifier/gmd:authority",
        )
        _authority = authority.make_config()
        if bool(_authority):
            _["authority"] = _authority

        code_value = self.record.xpath(
            f"{self.xpath}/gmd:referenceSystemInfo/gmd:MD_ReferenceSystem/gmd:referenceSystemIdentifier/"
            f"gmd:RS_Identifier/gmd:code/gco:CharacterString/text() | {self.xpath}/gmd:referenceSystemInfo/"
            f"gmd:MD_ReferenceSystem/gmd:referenceSystemIdentifier/gmd:RS_Identifier/gmd:code/gmx:Anchor/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(code_value) == 1:
            if "code" not in _:
                _["code"] = {}
            _["code"]["value"] = code_value[0]

        code_href = self.record.xpath(
            f"{self.xpath}/gmd:referenceSystemInfo/gmd:MD_ReferenceSystem/gmd:referenceSystemIdentifier/"
            f"gmd:RS_Identifier/gmd:code/gmx:Anchor/@xlink:href",
            namespaces=self.ns.nsmap(),
        )
        if len(code_href) == 1:
            _["code"]["href"] = code_href[0]

        version_value = self.record.xpath(
            f"{self.xpath}/gmd:referenceSystemInfo/gmd:MD_ReferenceSystem/gmd:referenceSystemIdentifier/"
            f"gmd:RS_Identifier/gmd:version/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(version_value) == 1:
            _["version"] = version_value[0]

        return _

    def make_element(self) -> None:
        reference_system_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}referenceSystemInfo")
        reference_system_element = SubElement(reference_system_wrapper, f"{{{self.ns.gmd}}}MD_ReferenceSystem")
        reference_system_identifier_wrapper = SubElement(
            reference_system_element, f"{{{self.ns.gmd}}}referenceSystemIdentifier"
        )
        reference_system_identifier_element = SubElement(
            reference_system_identifier_wrapper, f"{{{self.ns.gmd}}}RS_Identifier"
        )

        if "authority" in self.element_attributes:
            reference_system_identifier_authority_element = SubElement(
                reference_system_identifier_element, f"{{{self.ns.gmd}}}authority"
            )
            citation = Citation(
                record=self.record,
                attributes=self.attributes,
                parent_element=reference_system_identifier_authority_element,
                element_attributes=self.element_attributes["authority"],
            )
            citation.make_element()

        if "code" in self.element_attributes:
            reference_system_identifier_code_element = SubElement(
                reference_system_identifier_element, f"{{{self.ns.gmd}}}code"
            )
            if "href" in self.element_attributes["code"]:
                anchor = AnchorElement(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=reference_system_identifier_code_element,
                    element_attributes=self.element_attributes["code"],
                    element_value=self.element_attributes["code"]["value"],
                )
                anchor.make_element()
            else:
                reference_system_identifier_code_value = SubElement(
                    reference_system_identifier_code_element, f"{{{self.ns.gco}}}CharacterString"
                )
                reference_system_identifier_code_value.text = self.element_attributes["code"]["value"]

        if "version" in self.element_attributes:
            reference_system_identifier_version_element = SubElement(
                reference_system_identifier_element, f"{{{self.ns.gmd}}}version"
            )
            reference_system_identifier_version_value = SubElement(
                reference_system_identifier_version_element, f"{{{self.ns.gco}}}CharacterString"
            )
            reference_system_identifier_version_value.text = self.element_attributes["version"]



from __future__ import annotations

import json
import subprocess
from copy import deepcopy
from pathlib import Path
from tempfile import TemporaryDirectory

from importlib_resources import files as resource_file
from jsonschema import validate
from lxml.etree import (
    Element,
    ElementTree,
    fromstring,
)
from lxml.etree import (
    tostring as element_string,
)


class RecordValidationError(Exception):
    """Internal error indicating a record has failed schema validation."""

    pass


class Namespaces:
    """
    Gathers all XML namespaces used in a standard.

    Provides a way to reference namespaces when constructing elements and declaring namespaces in the root element of
    an XML document.

    This class is intended to be overridden in each metadata standard's module. See existing standards for examples.
    """

    _schema_locations = {}  # noqa: RUF012

    def __init__(self, namespaces: dict | None = None, root_namespace: str | None = None):
        """
        Root namespace support.

        The root namespace parameter is optional for use in standards that require a default namespace.

        When encoding data, this has the effect of not including a prefix in elements when encoded (e.g. `foo` instead
        of `ns:foo`). When decoding data, this behaviour must be suppressed using the relevant override parameter in
        the `nsmap()` method.

        @type namespaces: dict
        @param namespaces: dictionary of namespaces to add
        @type root_namespace: str
        @param root_namespace: optional namespace to use as unprefixed
        """
        self._namespaces = {}
        self.root_namespace = root_namespace
        if namespaces is not None:
            self._namespaces = {**self._namespaces, **namespaces}

    def nsmap(self, suppress_root_namespace: bool = False) -> dict:
        """
        Create a namespace map.

        Indexes namespaces by their prefix.

        E.g. {'xlink': 'http://www.w3.org/1999/xlink'}

        When a root namespace is set, a default namespace will be set by using the `None` constant for the relevant
        dict key (this is a lxml convention). This will create an invalid namespace map for use in XPath queries, this
        can be overcome using the `suppress_root_namespace` parameter, which will create a 'regular' map.

        :type suppress_root_namespace: bool
        :param suppress_root_namespace: When true, respects a root prefix as a default if set
        :return: dictionary of Namespaces indexed by prefix
        """
        nsmap = {}
        for prefix, namespace in self._namespaces.items():
            if hasattr(self, "root_namespace") and namespace == self.root_namespace and not suppress_root_namespace:
                nsmap[None] = namespace
                continue

            nsmap[prefix] = namespace

        return nsmap

    def schema_locations(self) -> str:
        """
        Generate value for a `xsi:schemaLocation` attribute.

        Defines the XML Schema Document (XSD) for each namespace in an XML tree

        E.g. 'xsi:schemaLocation="http://www.w3.org/1999/xlink https://www.w3.org/1999/xlink.xsd"'
        """
        schema_locations = ""
        for prefix, location in self._schema_locations.items():
            schema_locations = f"{schema_locations} {self._namespaces[prefix]} {location}"

        return schema_locations.lstrip()


class MetadataRecordConfig:
    """
    Represents the configuration for a metadata record.

    The record configuration can either be passed directly as a Python object (dict), when instantiating an instance of
    this class, or by using the `load()` or `loads()` methods to load the configuration from a JSON document.

    The structure and values of this configuration are specified by a JSON schema, and which should be used to validate
    configuration instances using the 'validate()' method. The `schema` property will hold the contents of that JSON
    Schema, the `schema_uri` lists the identity expected/required for that schema.

    It's expected this class will be used a base for standard specific classes, hard-coding the schema that must be
    used for validating a record configuration.
    """

    def __init__(self, **kwargs: dict):
        self.config: dict = kwargs
        self.schema: dict = {}
        self.schema_uri: str = ""

        if "$schema" not in self.config:
            self.config["$schema"] = self.schema_uri

    def validate(self) -> None:
        """
        Ensures the configuration is valid against the relevant JSON Schema.

        The record configuration (a Python dict) is first duplicated to a JSON safe encoding (i.e. Python dates objects
        converted to strings), to allow validation against the JSON Schema.

        Where the configuration is invalid, a relevant exception will be raised.
        """
        if self.schema is not None:
            _config = json.loads(json.dumps(deepcopy(self.config), default=str))
            validate(instance=_config, schema=self.schema)

    def load(self, file: Path) -> None:
        """Loads a record configuration from a JSON encoded file."""
        with file.open(mode="r") as file:
            self.config = json.load(fp=file)

    def loads(self, string: str) -> None:
        """Loads a record configuration from a JSON encoded string."""
        self.config = json.loads(s=string)

    def dump(self, file: Path) -> None:
        """Dumps a record configuration as a JSON encoded file."""
        with file.open(mode="w") as file:
            json.dump(self.config, file, indent=2)

    def dumps(self) -> str:
        """Dumps a record configuration as a JSON encoded string."""
        return json.dumps(self.config, indent=2)


class MetadataRecord:
    """
    Generates a metadata record using a configuration, or a configuration using a record.

    If a configuration is given, an XML tree of elements is built using the configuration object, typically for output
    as a complete record.

    If a record is given, the XML tree is parsed to reverse engineer a configuration object that describes it.

    These processes are designed to be lossless, meaning if a record is made from a configuration object, that record
    should be able to create exactly the same configuration object again without loosing any information.
    """

    def __init__(self, configuration: MetadataRecordConfig = None, record: str | None = None):
        self.ns = Namespaces()
        self.attributes = {}
        self.record = None

        if configuration is not None:
            configuration.validate()
            self.attributes = configuration.config

        if record is not None:
            self.record = fromstring(record.encode())

    def make_config(self) -> MetadataRecordConfig:
        """
        Builds a metadata configuration object by parsing an existing XML record.

        This method is effectively the reverse of `make_element()` and `generate_xml_document()`.
        """
        return MetadataRecordConfig(**self.attributes)

    def make_element(self) -> Element:
        """
        Builds a metadata record from a root XML element.

        Elements are added to this root element defining the contents of the record.
        """
        return None

    def generate_xml_document(self) -> bytes:
        """
        Generates an XML document and tree from an XML element defining a record.

        The XML document is encoded as a UTF-8 byte string, with pretty-printing and an XML declaration.
        """
        self.record = self.make_element()
        document = ElementTree(self.record)

        return element_string(document, pretty_print=True, xml_declaration=True, encoding="utf-8")

    def validate(self, xsd_path: Path) -> None:
        """
        Validates the contents of a record against a given XSD schema.

        The external `xmllint` binary is used to validate records as the `lxml` methods did not easily support relative
        paths for schemas that use imports/includes (which includes the ISO 19115 family).

        Schemas are loaded from an XSD directory within this package using the `importlib.files` method. The current
        record object is written to a temporary directory to easily pass to the `xmllint` binary.

        The `xmllint` binary only returns a 0 exit code if the record validates successfully. Therefore, any other exit
        code can be considered a validation failure, and returned as a `RecordValidationError` exception.

        It is assumed this method will be overridden in concrete implementations of this class. Specifically it's
        assumed the `xsd_path` parameter will be hard coded to a schema suitable for the standard each class implements.
        """
        schema_path = resource_file("bas_metadata_library.schemas.xsd").joinpath(xsd_path)

        with TemporaryDirectory() as document_path:
            document_path = Path(document_path).joinpath("record.xml")
            validation_document: MetadataRecord = deepcopy(self)
            with document_path.open(mode="w") as document_file:
                document_data = validation_document.generate_xml_document().decode()
                document_file.write(document_data)

            try:
                subprocess.run(
                    args=["xmllint", "--noout", "--schema", str(schema_path), str(document_path)],
                    capture_output=True,
                    check=True,
                )
            except subprocess.CalledProcessError as e:
                msg = f"Record validation failed: {e.stderr.decode()}"
                raise RecordValidationError(msg) from e


class MetadataRecordElement:
    """Create an XML element."""

    def __init__(
        self,
        record: Element,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        """
        Initialise.

        :type record: Element
        :param record: overall root element of a metadata record
        :type attributes: dict
        :param attributes: all attributes for a metadata record, from a record's configuration
        :type parent_element: Element
        :param parent_element: immediate parent of the current element
        :type element_attributes: dict
        :param element_attributes: attributes for the current element, taken from a record's configuration
        :type xpath: str
        :param xpath: Absolute XML XPath selecting the value of the element created
        """
        self.ns = Namespaces()
        self.record = record
        self.attributes = attributes
        self.parent_element = parent_element
        self.element_attributes = element_attributes
        self.xpath = xpath

        if self.parent_element is None:
            self.parent_element = self.record
        if self.element_attributes is None:
            self.element_attributes = self.attributes

    def make_config(self) -> None:
        """Parses an XML element to reverse engineer a partial configuration object."""
        pass

    def make_element(self) -> None:
        """Build an XML element."""
        pass



from __future__ import annotations

import json
from copy import deepcopy
from pathlib import Path

from importlib_resources import files as resource_file
from lxml.etree import Element, fromstring

from bas_metadata_library import MetadataRecord as _MetadataRecord
from bas_metadata_library import MetadataRecordConfig as _MetadataRecordConfig
from bas_metadata_library.standards.iso_19115_common import Namespaces
from bas_metadata_library.standards.iso_19115_common.root_element import ISOMetadataRecord
from bas_metadata_library.standards.iso_19115_common.utils import (
    decode_config_from_json,
    encode_config_for_json,
    validate_config,
)


class MetadataRecordConfigV4(_MetadataRecordConfig):
    """v4 configuration schema for ISO 19115:2003."""

    def __init__(self, **kwargs: dict):
        super().__init__(**kwargs)

        self.config = kwargs

        schema_path = resource_file("bas_metadata_library.schemas.dist").joinpath("iso_19115_2_v4.json")
        with schema_path.open() as schema_file:
            schema_data = json.load(schema_file)
        self.schema = schema_data

        # Workaround - will be addressed in #149
        self.schema_uri = schema_data["$id"]
        self.config = {"$schema": self.schema_uri, **kwargs}

    def validate(self) -> None:
        validate_config(config=self.config, schema=self.schema)

    def load(self, file: Path) -> None:
        with file.open() as file:
            self.config = decode_config_from_json(config=json.load(fp=file))

    def loads(self, string: str) -> None:
        self.config = decode_config_from_json(config=json.loads(s=string))

    def dump(self, file: Path) -> None:
        with file.open(mode="w") as file:
            json.dump(encode_config_for_json(config=deepcopy(self.config)), file, indent=2)

    def dumps(self) -> str:
        return json.dumps(encode_config_for_json(config=deepcopy(self.config)), indent=2)


class MetadataRecord(_MetadataRecord):
    """
    Defines the root element, and it's sub-elements, for this metadata standard.

    Expects/requires record configurations to use version 3 of the configuration schema for this standard
    """

    def __init__(self, configuration: MetadataRecordConfigV4 = None, record: str | None = None):
        self.ns = Namespaces()
        self.attributes = {}
        self.record = Element(
            f"{{{self.ns.gmi}}}MI_Metadata",
            attrib={f"{{{self.ns.xsi}}}schemaLocation": self.ns.schema_locations()},
            nsmap=self.ns.nsmap(),
        )
        self.xpath = "/gmi:MI_Metadata"

        if configuration is not None:
            configuration.validate()
            self.attributes = configuration.config

        if record is not None:
            self.record = fromstring(record.encode())

        self.metadata_record = ISOMetadataRecord(record=self.record, attributes=self.attributes, xpath=self.xpath)

    def make_config(self) -> MetadataRecordConfigV4:
        return MetadataRecordConfigV4(**self.metadata_record.make_config())

    def make_element(self) -> Element:
        return self.metadata_record.make_element()

    # noinspection PyMethodOverriding
    def validate(self) -> None:
        super().validate(xsd_path=Path("gmi/gmi.xsd"))
